import json
import shutil
from abc import ABCMeta, abstractmethod
from functools import cache
from threading import Lock
from typing import Optional

from ..state import VERSIONS, Version
from ..utils import *
from .base import Package


class Installer(metaclass=ABCMeta):
    @abstractmethod
    def install(self, package: str) -> None:
        pass

    @abstractmethod
    def upgrade(self, package: str) -> None:
        pass

    def is_installed(self, package: str) -> bool:
        return self.installed_version(package) is not None

    @abstractmethod
    def installed_version(self, package: str) -> Optional[str]:
        pass

    @abstractmethod
    def latest_version(self, package: str) -> str:
        pass


class Brew(Installer):
    def install(self, package: str) -> None:
        run("brew", "install", package)

    def upgrade(self, package: str) -> None:
        run("brew", "upgrade", package)

    @cache
    def info(self, package: str) -> tuple[str, Optional[str]]:
        info = json.loads(run_output("brew", "info", "--json=v2", package))
        if info["casks"]:
            cask = info["casks"][0]
            return cask["version"], cask["installed"]
        if info["formulae"]:
            formula = info["formulae"][0]
            try:
                installed = formula["installed"][0]["version"]
            except IndexError:
                installed = None
            version = formula["versions"]["stable"]
            revision = formula.get("revision")
            if revision:
                version += f"_{revision}"
            return version, installed
        raise ValueError(f"Unexpected output from brew: {info}")

    def installed_version(self, package: str) -> Optional[str]:
        try:
            return self.info(package)[1]
        except IndexError:
            return None

    def latest_version(self, package: str) -> str:
        return self.info(package)[0]


class DNF(Installer):
    def install(self, package: str) -> None:
        run("sudo", "dnf", "install", "-y", package)

    def upgrade(self, package: str) -> None:
        run("sudo", "dnf", "upgrade", "-y", package)

    def installed_version(self, package: str) -> Optional[str]:
        output = run_output(
            "rpm", "--query", "--queryformat", "%{VERSION}", "--whatprovides", package
        ).strip()
        if not output:
            return None
        return output

    def latest_version(self, package: str) -> str:
        output = run_output(
            "dnf",
            "--quiet",
            "repoquery",
            "--queryformat",
            "%{VERSION}",
            "--latest-limit",
            "1",
            "--arch",
            "x86_64,noarch",
            "--whatprovides",
            package,
        ).strip()
        if not output or "\n" in output:
            raise Exception(f"Cannot determine version for {package}.")
        return output


class Apt(Installer):
    def install(self, package: str) -> None:
        run("sudo", "apt", "install", "--yes", package)

    def upgrade(self, package: str) -> None:
        self.install(package)

    def latest_version(self, package: str) -> str:
        output = run_output(
            "apt-cache", "show", "--quiet", "--no-all-versions", package
        ).strip()
        for line in output.splitlines():
            line = line.strip()
            if line.startswith("Version:"):
                return line.split(": ", 1)[-1]
        raise Exception(f"Cannot determine version for {package}.")

    def installed_version(self, package: str) -> Optional[str]:
        try:
            return run_output(
                "dpkg-query", "--showformat", "${Version}", "--show", package
            ).strip()
        except subprocess.CalledProcessError:
            return None


def linux_installer() -> Installer:
    if shutil.which("dnf"):
        return DNF()
    elif shutil.which("apt"):
        return Apt()
    else:
        raise NotImplementedError("Cannot find a package manager.")


INSTALLER: Installer = with_os(
    linux=linux_installer, macos=lambda: Brew()  # pylint:disable=unnecessary-lambda
)()

INSTALLER_LOCK = Lock()


class SystemPackage(Package):
    def __init__(
        self,
        *,
        name: str,
        url: Optional[str] = None,
        auto_updates: bool = False,
        service: Some[str] = None,
        **kwargs,
    ) -> None:
        self._name = name
        self.url = url
        self.auto_updates = auto_updates
        self.services = unsome(service)
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return self._name

    def get_remote_version(self) -> str:
        if self.url:
            return url_version(self.url)
        if self.auto_updates:
            return "latest"
        return INSTALLER.latest_version(self.name)

    @property
    def local_version(self) -> Optional[str]:
        if self.url:
            try:
                return VERSIONS[self.name].version
            except KeyError:
                return None
        if self.auto_updates:
            return "latest" if INSTALLER.is_installed(self.name) else None
        return INSTALLER.installed_version(self.name)

    def postinstall_linux(self):
        if self.services:
            run("sudo", "systemctl", "daemon-reload")
            for service in self.services:
                run("sudo", "systemctl", "enable", service)

    def postinstall_macos(self):
        pass

    def install(self):
        with INSTALLER_LOCK:
            if self.url:
                INSTALLER.install(self.url)
                VERSIONS[self.name] = Version(version=self.get_remote_version())
            elif self.local_version:
                INSTALLER.upgrade(self.name)
            else:
                INSTALLER.install(self.name)
        with_os(linux=self.postinstall_linux, macos=self.postinstall_macos)()



import concurrent.futures
import os
import subprocess
import sys
from threading import Lock
from typing import Callable, Iterable, Literal, Optional, TypeVar, Union, cast

import requests
import tqdm  # type: ignore

OS = Literal["linux", "darwin"]

CURRENT_OS: OS = cast(OS, sys.platform)


Distribution = str


def get_distro() -> Optional[Distribution]:
    if CURRENT_OS != "linux":
        return None

    release_file = "/etc/os-release"
    with open(release_file) as release:
        for line in release:
            k, v = line.strip().split("=", 1)
            if k == "ID":
                return v

    raise ValueError(f"Cannot find distribution ID in {release_file}.")


CURRENT_DISTRIBUTION = get_distro()

T = TypeVar("T")
U = TypeVar("U")


def with_os(*, linux: T, macos: T) -> T:
    if CURRENT_OS == "linux":
        return linux
    elif CURRENT_OS == "darwin":
        return macos
    else:
        raise ValueError(f"Unexpected OS {CURRENT_OS}.")


Some = Optional[Union[T, list[T]]]


def unsome_(x: Some[T]) -> Optional[list[T]]:
    if x is None:
        return None
    if isinstance(x, list):
        return x
    return [x]


def unsome(x: Some[T]) -> list[T]:
    return unsome_(x) or []


TERMINAL_LOCK = Lock()


def run(*args: str, **kwargs) -> subprocess.CompletedProcess:
    if args[0] == "sudo":
        # If the sudo prompt is needed, avoid drawing a progress bar over it
        # with first prompting for a no-op command.
        with TERMINAL_LOCK:
            subprocess.run(["sudo", "true"], check=True)
    return subprocess.run(args, check=True, **kwargs)


def run_ok(*args: str, **kwargs) -> bool:
    return (
        subprocess.run(
            args,
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            **kwargs,
        ).returncode
        == 0
    )


def run_output(*args: str, **kwargs) -> str:
    return (
        subprocess.run(args, stdout=subprocess.PIPE, check=True, **kwargs)
        .stdout.decode()
        .strip()
    )


def home(*path: str) -> str:
    return os.path.join(os.environ["HOME"], *path)


def local(*path: str) -> str:
    return home(".local", *path)


def flatten(items: Iterable[Iterable[T]]) -> list[T]:
    return [item for sublist in items for item in sublist]


def parallel_map_tqdm(items: list[T], action: Callable[[T], U]) -> list[U]:
    with tqdm.tqdm(total=len(items)) as progress:
        with concurrent.futures.ThreadPoolExecutor(20) as executor:

            def action_and_update(item: T) -> U:
                result = action(item)
                with TERMINAL_LOCK:
                    progress.update(1)
                return result

            return list(executor.map(action_and_update, items))


class Filters:
    @staticmethod
    def includes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring in x.lower()

    @staticmethod
    def excludes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring not in x

    @staticmethod
    def startswith(prefix: str) -> Callable[[str], bool]:
        return lambda x: x.startswith(prefix)

    @staticmethod
    def endswith(suffix: str) -> Callable[[str], bool]:
        return lambda x: x.endswith(suffix)


def url_version(url: str) -> str:
    head_response = requests.head(url, allow_redirects=True)
    return head_response.headers["etag"]



import os
from contextlib import contextmanager
from typing import Iterator, Optional

from ..utils import *
from .base import Package


class Clone(Package):
    def __init__(self, clone: str, destination: str, **kwargs) -> None:
        self.repo = clone
        self.destination = destination
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return self.repo

    @property
    def directory(self) -> str:
        return home(self.destination)

    @property
    def directory_exists(self) -> bool:
        return os.path.isdir(self.directory)

    @contextmanager
    def in_directory(self) -> Iterator[None]:
        olddir = os.getcwd()
        try:
            os.chdir(self.directory)
            yield
        finally:
            os.chdir(olddir)

    @property
    def local_version(self) -> Optional[str]:
        if not self.directory_exists:
            return None
        with self.in_directory():
            return run_output("git", "rev-parse", "HEAD")

    @property
    def remote(self):
        return f"https://github.com/{self.repo}.git"

    def get_remote_version(self) -> str:
        return run_output("git", "ls-remote", self.remote, "HEAD").split()[0]

    def install(self) -> None:
        if not self.directory_exists:
            run("git", "clone", self.remote, self.directory)
        with self.in_directory():
            run("git", "remote", "set-url", "origin", self.remote)
            run("git", "fetch")
            default_branch = run_output(
                "git", "rev-parse", "--abbrev-ref", "origin/HEAD"
            ).split("/")[1]
            run("git", "switch", default_branch)
            run("git", "reset", "--hard", f"origin/{default_branch}")



import hashlib
import os
from typing import Iterator

import yaml

from ..fs import files_in_recursively, link
from ..utils import *
from .manual_version import ManualVersion


class Links(ManualVersion):
    def __init__(
        self,
        links: str,
        dest: str,
        dot: bool = False,
        root: bool = False,
        shallow: bool = False,
        only: Some[str] = None,
        **kwargs,
    ) -> None:
        self.root = root
        self.source = os.path.abspath(links)
        if not dest.startswith("/"):
            if self.root:
                dest = os.path.join(with_os(linux="/root", macos="/var/root"), dest)
            else:
                dest = os.path.join(home(), dest)
        self.dest = dest
        self.dot = dot
        self.shallow = shallow
        self.only = unsome_(only)
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return f"links-{self.source}-{self.dest}"

    def all_paths(self) -> Iterator[str]:
        if self.shallow:
            for entry in os.scandir(self.source):
                yield entry.path
        else:
            for path in files_in_recursively(self.source):
                if os.path.isfile(path):
                    yield path

    def paths(self) -> Iterator[str]:
        for path in self.all_paths():
            if self.only and os.path.basename(path) not in self.only:
                continue
            yield path

    def get_remote_version(self) -> str:
        m = hashlib.sha256()
        for path in self.paths():
            m.update(path.encode())
        return m.hexdigest()

    def install(self) -> None:
        # TODO: remove links that were created before but no longer exist
        for path in self.paths():
            target = os.path.relpath(path, self.source)
            if self.dot:
                target = "." + target
            target = os.path.join(self.dest, target)

            link(path, target, sudo=self.root)
        super().install()


def load_links() -> list[Links]:
    with open("links.yaml") as f:
        links = yaml.safe_load(f)
        return list(Links(**args) for args in links)



import json
import subprocess
from typing import Optional

from ..utils import run, run_output
from .pip_base import PipBasePackage


def pip_cmd(cmd: str, /, *args: str) -> list[str]:
    return ["python3", "-m", "pip", cmd, "--user", *args]


def get_pip_versions() -> dict[str, str]:
    packages = json.loads(
        run_output(*pip_cmd("list", "--format", "json"), stderr=subprocess.DEVNULL)
    )
    return {package["name"]: package["version"] for package in packages}


class PipPackage(PipBasePackage):
    def __init__(self, *, pip: str, **kwargs) -> None:
        self.package = pip
        super().__init__(**kwargs)

    @property
    def local_version(self) -> Optional[str]:
        return get_pip_versions().get(self.package)

    def install(self) -> None:
        cmd = "install" if self.local_version is None else "upgrade"
        run(*pip_cmd(cmd, self.package))



import argparse

from .package import load_packages
from .utils import flatten, parallel_map_tqdm


def parser() -> argparse.ArgumentParser:
    result = argparse.ArgumentParser(description="Install the environment")
    result.add_argument("component", nargs="*", help="The component to install")
    return result


def main():
    args = parser().parse_args()
    components: frozenset[str] = frozenset(args.component) | {"base"}
    packages = flatten(map(load_packages, components))
    results = parallel_map_tqdm(packages, lambda p: p.ensure())
    total_installed = sum(1 if installed else 0 for installed in results)
    print(f"{total_installed} packages installed or updated.")



import json
import subprocess
from typing import Optional

from ..utils import run, run_output
from .pip_base import PipBasePackage

PIPX = ["python3", "-m", "pipx"]


def get_pipx_versions() -> dict[str, str]:
    pipx_list = json.loads(
        run_output(*PIPX, "list", "--json", stderr=subprocess.DEVNULL)
    )
    packages = (
        item["metadata"]["main_package"] for item in pipx_list["venvs"].values()
    )
    return {package["package"]: package["package_version"] for package in packages}


class PipxPackage(PipBasePackage):
    def __init__(self, *, pipx: str, **kwargs) -> None:
        self.package = pipx
        super().__init__(**kwargs)

    @property
    def local_version(self) -> Optional[str]:
        return get_pipx_versions().get(self.package)

    def install(self) -> None:
        cmd = "install" if self.local_version is None else "upgrade"
        run(*PIPX, cmd, self.package)
